/*
 * Copyright (c) 2013-2017 ARM Limited. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "Driver_ETH_MAC.h"
#include "Driver_ETH_PHY.h"
#include "rpmsg_rtos.h"
#include <string.h>

#define ARM_ETH_MAC_DRV_VERSION    ARM_DRIVER_VERSION_MAJOR_MINOR(1, 0) /* driver version */
#define ARM_ETH_PHY_DRV_VERSION    ARM_DRIVER_VERSION_MAJOR_MINOR(1, 0) /* driver version */

/* RPMSG Ethernet endpoint */
#ifndef ETHERNET_ENDPT
#define ETHERNET_ENDPT       125U
#endif

/* Memory Buffer configuration */
#ifndef ETH_BUF_SIZE
#define ETH_BUF_SIZE        1536U
#endif

/* MAC Driver Version */
static const ARM_DRIVER_VERSION MAC_DriverVersion = {
  ARM_ETH_MAC_API_VERSION,
  ARM_ETH_MAC_DRV_VERSION
};

/* PHY Driver Version */
static const ARM_DRIVER_VERSION PHY_DriverVersion = {
  ARM_ETH_PHY_API_VERSION,
  ARM_ETH_PHY_DRV_VERSION
};

/* Driver Capabilities */
static const ARM_ETH_MAC_CAPABILITIES DriverCapabilities = {
  0U,    /* 1 = IPv4 header checksum verified on receive */
  0U,    /* 1 = IPv6 checksum verification supported on receive */
  0U,    /* 1 = UDP payload checksum verified on receive */
  0U,    /* 1 = TCP payload checksum verified on receive */
  0U,    /* 1 = ICMP payload checksum verified on receive */
  0U,    /* 1 = IPv4 header checksum generated on transmit */
  0U,    /* 1 = IPv6 checksum generation supported on transmit */
  0U,    /* 1 = UDP payload checksum generated on transmit */
  0U,    /* 1 = TCP payload checksum generated on transmit */
  0U,    /* 1 = ICMP payload checksum generated on transmit */
  0U,    /* Ethernet Media Interface type */
  0U,    /* 1 = driver provides initial valid MAC address */
  0U,    /* 1 = callback event \ref ARM_ETH_MAC_EVENT_RX_FRAME generated */
  0U,    /* 1 = callback event \ref ARM_ETH_MAC_EVENT_TX_FRAME generated */
  0U,    /* 1 = wakeup event \ref ARM_ETH_MAC_EVENT_WAKEUP generated */
  0U     /* 1 = Precision Timer supported */
};

/* Local variables */
static struct remote_device *rdev = NULL;
static struct rpmsg_channel *app_chnl = NULL;
static struct rpmsg_endpoint *eptEthernet;

static ARM_ETH_MAC_ADDR mac_addr;
static uint32_t len_frame = 0U;
static uint8_t buffer[ETH_BUF_SIZE];

/* Ethernet MAC Driver functions */

/**
  \fn          ARM_DRIVER_VERSION GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRIVER_VERSION
*/
ARM_DRIVER_VERSION ARM_ETH_MAC_GetVersion(void)
{
  return MAC_DriverVersion;
}

/**
  \fn          ARM_ETH_MAC_CAPABILITIES GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_ETH_MAC_CAPABILITIES
*/
ARM_ETH_MAC_CAPABILITIES ARM_ETH_MAC_GetCapabilities(void)
{
  return DriverCapabilities;
}

/**
  \fn          int32_t Initialize (ARM_ETH_MAC_SignalEvent_t cb_event)
  \brief       Initialize Ethernet MAC Device.
  \param[in]   cb_event  Pointer to \ref ARM_ETH_MAC_SignalEvent
  \return      \ref execution_status
*/
int32_t ARM_ETH_MAC_Initialize(ARM_ETH_MAC_SignalEvent_t cb_event)
{
  rpmsg_rtos_init(0 /*REMOTE_CPU_ID*/, &rdev, RPMSG_MASTER, &app_chnl);
  eptEthernet = rpmsg_rtos_create_ept(app_chnl, ETHERNET_ENDPT);

  if(eptEthernet == NULL){
    return ARM_DRIVER_ERROR;
  }

  mac_addr.b[5U] = (uint8_t)0U;
  mac_addr.b[4U] = (uint8_t)0U;
  mac_addr.b[3U] = (uint8_t)0U;
  mac_addr.b[2U] = (uint8_t)0U;
  mac_addr.b[1U] = (uint8_t)0U;
  mac_addr.b[0U] = (uint8_t)0U;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t Uninitialize (void)
  \brief       De-initialize Ethernet MAC Device.
  \return      \ref execution_status
*/
int32_t ARM_ETH_MAC_Uninitialize(void)
{
  rpmsg_rtos_deinit(rdev);
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t PowerControl (ARM_POWER_STATE state)
  \brief       Control Ethernet MAC Device Power.
  \param[in]   state  Power state
  \return      \ref execution_status
*/
int32_t ARM_ETH_MAC_PowerControl(ARM_POWER_STATE state)
{
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t GetMacAddress (ARM_ETH_MAC_ADDR *ptr_addr)
  \brief       Get Ethernet MAC Address.
  \param[in]   ptr_addr  Pointer to address
  \return      \ref execution_status
*/
int32_t ARM_ETH_MAC_GetMacAddress(ARM_ETH_MAC_ADDR *ptr_addr)
{
  if (!ptr_addr) {
    /* Invalid parameters */
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  ptr_addr->b[5U] = mac_addr.b[5U];
  ptr_addr->b[4U] = mac_addr.b[4U];
  ptr_addr->b[3U] = mac_addr.b[3U];
  ptr_addr->b[2U] = mac_addr.b[2U];
  ptr_addr->b[1U] = mac_addr.b[1U];
  ptr_addr->b[0U] = mac_addr.b[0U];
  
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t SetMacAddress (const ARM_ETH_MAC_ADDR *ptr_addr)
  \brief       Set Ethernet MAC Address.
  \param[in]   ptr_addr  Pointer to address
  \return      \ref execution_status
*/
int32_t ARM_ETH_MAC_SetMacAddress(const ARM_ETH_MAC_ADDR *ptr_addr)
{
  if (!ptr_addr) {
    /* Invalid parameters */
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  /* Set Ethernet MAC Address */
  mac_addr.b[5U] = ptr_addr->b[5U];
  mac_addr.b[4U] = ptr_addr->b[4U];
  mac_addr.b[3U] = ptr_addr->b[3U];
  mac_addr.b[2U] = ptr_addr->b[2U];
  mac_addr.b[1U] = ptr_addr->b[1U];
  mac_addr.b[0U] = ptr_addr->b[0U];

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t SetAddressFilter (const ARM_ETH_MAC_ADDR *ptr_addr,
                                               uint32_t          num_addr)
  \brief       Configure Address Filter.
  \param[in]   ptr_addr  Pointer to addresses
  \param[in]   num_addr  Number of addresses to configure
  \return      \ref execution_status
*/
int32_t ARM_ETH_MAC_SetAddressFilter(const ARM_ETH_MAC_ADDR *ptr_addr, uint32_t num_addr)
{
  return ARM_DRIVER_ERROR_UNSUPPORTED;
}

/**
  \fn          int32_t SendFrame (const uint8_t *frame, uint32_t len, uint32_t flags)
  \brief       Send Ethernet frame.
  \param[in]   frame  Pointer to frame buffer with data to send
  \param[in]   len    Frame buffer length in bytes
  \param[in]   flags  Frame transmit flags (see ARM_ETH_MAC_TX_FRAME_...)
  \return      \ref execution_status
*/
int32_t ARM_ETH_MAC_SendFrame(const uint8_t *frame, uint32_t len, uint32_t flags)
{
  int rcMsg;
  if (!frame || !len) {
  /* Invalid parameters */
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  rcMsg = rpmsg_rtos_send(eptEthernet, (void *)frame, len, ETHERNET_ENDPT);
  if (rcMsg != NULL) {
    return ARM_DRIVER_ERROR;
  } else {
    return ARM_DRIVER_OK;
  }
}

/**
  \fn          int32_t ReadFrame (uint8_t *frame, uint32_t len)
  \brief       Read data of received Ethernet frame.
  \param[in]   frame  Pointer to frame buffer for data to read into
  \param[in]   len    Frame buffer length in bytes
  \return      number of data bytes read or execution status
                 - value >= 0: number of data bytes read
                 - value < 0: error occurred, value is execution status as defined with \ref execution_status 
*/
int32_t ARM_ETH_MAC_ReadFrame(uint8_t *frame, uint32_t len)
{
  uint32_t ret = 0U;

  if (!frame && len) {
    /* Invalid parameters */
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  if (len_frame != 0U) {
    memcpy(frame, buffer, len);
    ret = len_frame;
    len_frame = 0U;
  }

  return (ret);
}

/**
  \fn          uint32_t GetRxFrameSize (void)
  \brief       Get size of received Ethernet frame.
  \return      number of bytes in received frame
*/
uint32_t ARM_ETH_MAC_GetRxFrameSize(void)
{
  int bytesRx = 0U;

  if (len_frame == 0U){
    rpmsg_rtos_recv(eptEthernet, &buffer, &bytesRx, ETH_BUF_SIZE, NULL, 0U);
    len_frame = bytesRx;
  }
  
  return (len_frame);
}

/**
  \fn          int32_t GetRxFrameTime (ARM_ETH_MAC_TIME *time)
  \brief       Get time of received Ethernet frame.
  \param[in]   time  Pointer to time structure for data to read into
  \return      \ref execution_status
*/
int32_t ARM_ETH_MAC_GetRxFrameTime(ARM_ETH_MAC_TIME *time)
{
  return 0U;
}

/**
  \fn          int32_t GetTxFrameTime (ARM_ETH_MAC_TIME *time)
  \brief       Get time of transmitted Ethernet frame.
  \param[in]   time  Pointer to time structure for data to read into
  \return      \ref execution_status
*/
int32_t ARM_ETH_MAC_GetTxFrameTime(ARM_ETH_MAC_TIME *time)
{
  return 0U;
}

/**
  \fn          int32_t ControlTimer (uint32_t control, ARM_ETH_MAC_TIME *time)
  \brief       Control Precision Timer.
  \param[in]   control  Operation
  \param[in]   time     Pointer to time structure
  \return      \ref execution_status
*/
int32_t ARM_ETH_MAC_Control(uint32_t control, uint32_t arg)
{
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t Control (uint32_t control, uint32_t arg)
  \brief       Control Ethernet Interface.
  \param[in]   control  Operation
  \param[in]   arg      Argument of operation (optional)
  \return      \ref execution_status
*/
int32_t ARM_ETH_MAC_ControlTimer(uint32_t control, ARM_ETH_MAC_TIME *time)
{
  return 0U;
}

/**
  \fn          int32_t PHY_Read (uint8_t phy_addr, uint8_t reg_addr, uint16_t *data)
  \brief       Read Ethernet PHY Register through Management Interface.
  \param[in]   phy_addr  5-bit device address
  \param[in]   reg_addr  5-bit register address
  \param[out]  data      Pointer where the result is written to
  \return      \ref execution_status
*/
int32_t ARM_ETH_MAC_PHY_Read(uint8_t phy_addr, uint8_t reg_addr, uint16_t *data)
{
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t PHY_Write (uint8_t phy_addr, uint8_t reg_addr, uint16_t data)
  \brief       Write Ethernet PHY Register through Management Interface.
  \param[in]   phy_addr  5-bit device address
  \param[in]   reg_addr  5-bit register address
  \param[in]   data      16-bit data to write
  \return      \ref execution_status
*/
int32_t ARM_ETH_MAC_PHY_Write(uint8_t phy_addr, uint8_t reg_addr, uint16_t data)
{
  return ARM_DRIVER_OK;
}

/* MAC Driver Control Block */
ARM_DRIVER_ETH_MAC Driver_ETH_MAC0 =
{
  ARM_ETH_MAC_GetVersion,
  ARM_ETH_MAC_GetCapabilities,
  ARM_ETH_MAC_Initialize,
  ARM_ETH_MAC_Uninitialize,
  ARM_ETH_MAC_PowerControl,
  ARM_ETH_MAC_GetMacAddress,
  ARM_ETH_MAC_SetMacAddress,
  ARM_ETH_MAC_SetAddressFilter,
  ARM_ETH_MAC_SendFrame,
  ARM_ETH_MAC_ReadFrame,
  ARM_ETH_MAC_GetRxFrameSize,
  ARM_ETH_MAC_GetRxFrameTime,
  ARM_ETH_MAC_GetTxFrameTime,
  ARM_ETH_MAC_ControlTimer,
  ARM_ETH_MAC_Control,
  ARM_ETH_MAC_PHY_Read,
  ARM_ETH_MAC_PHY_Write
};


/* Ethernet PHY Driver functions */

/**
  \fn          ARM_DRIVER_VERSION GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRIVER_VERSION
*/
ARM_DRIVER_VERSION ARM_ETH_PHY_GetVersion(void)
{
  return PHY_DriverVersion;
}

/**
  \fn          int32_t Initialize (ARM_ETH_PHY_Read_t  fn_read,
                                   ARM_ETH_PHY_Write_t fn_write)
  \brief       Initialize Ethernet PHY Device.
  \param[in]   fn_read   Pointer to \ref ARM_ETH_MAC_PHY_Read
  \param[in]   fn_write  Pointer to \ref ARM_ETH_MAC_PHY_Write
  \return      \ref execution_status
*/
int32_t ARM_ETH_PHY_Initialize(ARM_ETH_PHY_Read_t fn_read, ARM_ETH_PHY_Write_t fn_write)
{
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t Uninitialize (void)
  \brief       De-initialize Ethernet PHY Device.
  \return      \ref execution_status
*/
int32_t ARM_ETH_PHY_Uninitialize(void)
{
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t PowerControl (ARM_POWER_STATE state)
  \brief       Control Ethernet PHY Device Power.
  \param[in]   state  Power state
  \return      \ref execution_status
*/
int32_t ARM_ETH_PHY_PowerControl(ARM_POWER_STATE state)
{
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t SetInterface (uint32_t interface)
  \brief       Set Ethernet Media Interface.
  \param[in]   interface  Media Interface type
  \return      \ref execution_status
*/
int32_t ARM_ETH_PHY_SetInterface(uint32_t interface)
{
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t SetMode (uint32_t mode)
  \brief       Set Ethernet PHY Device Operation mode.
  \param[in]   mode  Operation Mode
  \return      \ref execution_status
*/
int32_t ARM_ETH_PHY_SetMode(uint32_t mode)
{
  return ARM_DRIVER_OK;
}

/**
  \fn          ARM_ETH_LINK_STATE GetLinkState (void)
  \brief       Get Ethernet PHY Device Link state.
  \return      current link status \ref ARM_ETH_LINK_STATE
*/
ARM_ETH_LINK_STATE ARM_ETH_PHY_GetLinkState(void)
{
  return ARM_ETH_LINK_UP;
}

/**
  \fn          ARM_ETH_LINK_INFO GetLinkInfo (void)
  \brief       Get Ethernet PHY Device Link information.
  \return      current link parameters \ref ARM_ETH_LINK_INFO
*/
ARM_ETH_LINK_INFO ARM_ETH_PHY_GetLinkInfo(void)
{
  ARM_ETH_LINK_INFO info;
  info.duplex = 1;
  info.speed = 0;
  return info;
}

/* PHY Driver Control Block */
ARM_DRIVER_ETH_PHY Driver_ETH_PHY0 =
{
  ARM_ETH_PHY_GetVersion,
  ARM_ETH_PHY_Initialize,
  ARM_ETH_PHY_Uninitialize,
  ARM_ETH_PHY_PowerControl,
  ARM_ETH_PHY_SetInterface,
  ARM_ETH_PHY_SetMode,
  ARM_ETH_PHY_GetLinkState,
  ARM_ETH_PHY_GetLinkInfo,
};
